/*********************************************************************************************/
%
% internalStructures.pl (Stundenplangenerierung)
%
% Prdikate zum Aufbau der internen Strukturen aus den Informationen der externen           
% Datenbasis.                                                                               
%                                                                                           
/*********************************************************************************************/

:- use_module(library(fd)).
	
/*************************************************************************/
%
% blockNumberList/2(+CourseList,-BNumbers)
%
% Liest die Vorlesungsdefinitionen aus der vorlesung/4-Liste CourseList,
% und bildet aus den Blockdefinitionen jeder Vorlesung die Liste der
% Blcke BNumbers
%
% +CourseList : Die von der Proc. preparedCourses erstellte vorlesung/4-Liste 
%               der Vorlesungen.
% -BNumbers   : block/3-Liste aller Blcke des Stundenplanes. Die Liste besteht
%               aus block/3(BID,VName,Struct) Strukturen.
%                   BID   : eindeutige ID des Blockes (Integer)
%                   VName : Name der Vorlesung, zu der der Block gehrt
%                   Struct: Element aus {single,singleNoCluster,  
%                                        block2,block2NoCluster,
%					 block3,block3NoCluster}
%                           Definiert welche Grsse der Block hat, und ob er
%                           gemeinsam mit anderen Blcken derselben Vorlesung
%                           an einem Tag liegen darf (xxxNoCluster-Blcke 
%                           drfen das nicht)
%
%*************************************************************************/
blockNumberList(PreparedCourses,BNumbers) :-
    blockNumberList(PreparedCourses,[],BNumbers,0).

% Die eigentliche rekursive Proc. 
% blockNumberList/3(CourseList,Akkumulator,Result,ActBlockID)
blockNumberList([],Result,Result,_).
blockNumberList([vorlesung(Name,_,_,BlockList)|PCT],AccResult,BNumbers,ActBlockID) :-
    processBlockList(Name,ActBlockID,NewBlockID,BlockList,BList),
    append(BList,AccResult,NewResult),
    blockNumberList(PCT,NewResult,BNumbers,NewBlockID).

processBlockList(_,BID,BID,[],[]).
processBlockList(CourseName,ActBlockID,NewBlockID,[single(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,1,N,single,NewBlockID,Result).
processBlockList(CourseName,ActBlockID,NewBlockID,[singleNoCluster(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,1,N,singleNoCluster,NewBlockID,Result).   
processBlockList(CourseName,ActBlockID,NewBlockID,[block2(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,2,N,block2,NewBlockID,Result). 
processBlockList(CourseName,ActBlockID,NewBlockID,[block2NoCluster(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,2,N,block2NoCluster,NewBlockID,Result).
processBlockList(CourseName,ActBlockID,NewBlockID,[block3(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,3,N,block3,NewBlockID,Result).
processBlockList(CourseName,ActBlockID,NewBlockID,[block3NoCluster(N)|BLT],Result) :-
    processBlockStructure(ActBlockID,CourseName,BLT,3,N,block3NoCluster,NewBlockID,Result).

processBlockStructure(ActBlockID,CourseName,BListRest,BlockSize,N,BlockStructure,NewBlockID,Result) :-
     NextBlockID is ActBlockID+N, 
     processBlockList(CourseName,NextBlockID,NewBlockID,BListRest,Result0),
     getNBlocksOfSize(N,BlockSize,ActBlockID,CourseName,BlockStructure,BlockList),
     append(BlockList,Result0,Result).

getNBlocksOfSize(0,_,_,_,_,[]) :- !.
getNBlocksOfSize(N,Size,ActBlockID,CourseName,BlockStruct,BList) :-
    M is N-1,
    NextID is ActBlockID+1,
    getNBlocksOfSize(M,Size,NextID,CourseName,BlockStruct,Result0),
    append([block(ActBlockID,CourseName,BlockStruct)],Result0,BList).
    
/*************************************************************************/
%
% preparedCourseList/1(-PreparedCourses)
%
% Liefert eine Liste von vorlesung/4 Strukturen, die nach Kursgroesse absteigend
% sortiert ist. Die Informationen ber definierte Vorlesungen erhlt diese Prozedur 
% aus der Datenbasis. Vorlesungen die als terminNachAbsprache/1 definiert sind 
% werden ignoriert. Es werden alle Vorlesung/4 Strukturen aus der Datenbasis 
% geholt, und in die Liste -PreparedCourses bertragen, wobei der Resourcenliste
% jeder Vorlesung die Struktur groesse/1 hinzugefgt wird, die als Parameter die
% aus den Gruppendefinitionen aufsummierte Gesamtteilnehmerzahl enthlt. Der
% Liste der Teilnehmergruppen werden alle zu den enthaltenen Gruppen nicht disjunkten 
% Gruppen hinzugefgt (Dies wird z.b. durch Subgruppendefinitionen ntig)
%
% -PreparedCourses : Liste aller vorlesung/4 Strukturen aus der DB, die nicht als
%                    terminNachAbsprache/1 definiert sind.
%                       vorlesung/4(VName,Gruppen,Resourcen,Blockungen)
%                          VName     : Vorlesungsname
%                          Gruppen   : Wie in DB definiert + alle dazu nicht disjunkten
%                          Resourcen : -- " " ------------ + groesse/1 - Struktur
%                          Blockungen: -- " " ------------
%
%*************************************************************************/
preparedCourseList(PreparedCourses) :-
	findall(SumOfListeners-vorlesung(Name,NewGroups,NewResources,Blocks),
	  (vorlesung(Name,Groups,Resources,Blocks),
	   not(terminNachAbsprache(Name)),
	   sumAllListeners(Groups,SumOfListeners),
	   append([groesse(SumOfListeners)],Resources,NewResources),
	   addNonDisjointGroups(Groups,NewGroups)
	  ),
	  AllCourses),
	keysort(AllCourses,TempCourses),
	reverse(TempCourses,AllCoursesSorted),
	findall(vorlesung(N,G,R,B),member(_-vorlesung(N,G,R,B),AllCoursesSorted),PreparedCourses).

addNonDisjointGroups(Groups,NewGroups) :-
    addNonDisjointGroups(Groups,NewGroups0,[]),
    append(NewGroups0,Groups,NewGroups1),
    removeDoubles(NewGroups1,NewGroups).

addNonDisjointGroups([],NewGroups,NewGroups).
addNonDisjointGroups([GroupsH|T],NewGroups,AccRes) :-
    addAllNonDisjoint(GroupsH,AccRes,NewRes),
    addNonDisjointGroups(T,NewGroups,NewRes).

addAllNonDisjoint(Group,OtherGroups,Res):-
    group(Group,_),!,
    findall(SubGroupName,subgroup(SubGroupName,Group,_,_),SubGroups),
    append(SubGroups,OtherGroups,Res).
addAllNonDisjoint(Group,OtherGroups,Res):-
    subgroup(Group,Parent,A,_),!,
    findall(Other,(subgroup(Other,Parent,B,_),A\=B),Others),
    append(Others,[Parent],Temp),
    append(Temp,OtherGroups,Res).    
addAllNonDisjoint(_,Res,Res).
    
    

    
% Summiert die Grssen aller Hrergruppen. Die bergebene Liste ist eine Group-Liste 
% einer Vorlesung. Wird von preparedCourseList benutzt 
%   sumAllListeners/2(+Groups,-Size)
%     Groups : Liste von Gruppennamen.
%     Size   : Aufsummierte Gre aller Gruppen aus Groups
sumAllListeners([],0).
sumAllListeners([H|T],Sum) :-
   dozent(H),!,
   sumAllListeners(T,Sum).
sumAllListeners([H|T],Sum) :-
   group(H,Size),!,
   sumAllListeners(T,Sum0),
   Sum is Sum0 + Size.
sumAllListeners([H|T],Sum) :-
   subgroup(H,_,_,Size),!,
   sumAllListeners(T,Sum0),
   Sum is Sum0 + Size.
sumAllListeners([H|T],Sum) :-
   message(['Nicht definierte Gruppe gefunden -> ',H]),
   sumAllListeners(T,Sum).


/*************************************************************************/
%
% notTheSameTime/2(+CourseList,-CoursePairs)
%
% Bildet die ungeordneten Paare aller Vorlesungen die nicht gleichzeitig
% stattfinden drfen, da ihre Teilnehmergruppen nicht disjunkt sind.
% Wird von prepareInternals/0 benutzt. Die ErgebnisListe dieser Proc. wird
% von der Proc. prepareInternals/0 als commonGroups/1 in die Datenbasis
% geschrieben.
%
% +CourseList : Liste von vorlesung/4 Strukturen
% -CoursePairs: Liste aller Paare nst/2(VNameA,VNameB) von Vorlesungen die
%               nicht gleichzeitig stattfinden drfen. VNameA bzw. B sind die
%               Namen der Vorlesungen.
%
%*************************************************************************/
notTheSameTime([],[]).
notTheSameTime([vorlesung(A,Groups_1,_,_)|PCT],CoursePairs) :-
   findall(
     nst(A,B),
     (
       member(vorlesung(B,Groups_2,_,_),PCT),
       groupsAreNotDisjoint(Groups_1,Groups_2)
     ),
     CoursePairs0
   ),
   notTheSameTime(PCT,CoursePairs1),
   append(CoursePairs0,CoursePairs1,CoursePairs).
   
%  Wahr wenn die Schnittmenge Elemente der beiden bergebenen Listen  
%  nicht leer ist.                                                    
groupsAreNotDisjoint(Groups1,Groups2) :-
	member(Gr,Groups1),
	member(Gr,Groups2),
	!.

/*************************************************************************/
%
% mkVariables/2(+List,-Variables) 
%
%  Erstellt eine Liste aus dtr(Day,Time,Room) Strukturen deren Kardinalitt der 
%  Anzahl aller Zeiteinheiten aller Veranstaltungen entspricht. Die freien 
%  Variable der dtr/4(Tag,Zeit,Raum,BlockID) Strukturen sind mit Ausnahme von BlockID 
%  Domain-Variable mit entsprechenden Wertebereichen. Die zurckgelieferte Variablenliste 
%  bildet den Suchraum des Stundenplanproblems.
%
%  +List  : Liste aller Blcke block/3(BID,VName,Struct) aller Vorlesungen, wie
%           sie von blockNumberList/2 erstellt wurde.
%
%  -Variables : Liste aller dtr/4(Tag,Zeit,Raum,BlockID) Variablenstrukturen. 
%               Jede Struktur mit ihren Variablen entspricht dabei einer Zeiteinheit
%               an einem Tag im Stundenplan, und ist Teil eines Blockes einer Vorlesung.
%               Tag,Zeit,Raum sind ConstraintVariable mit entsprechenden Wertebereichen.
%
%*************************************************************************/
mkVariables(Blocks,Variables) :-
	noOfRooms(NR),
	noOfDays(ND),
	noOfTimes(NT),
	noOfTimesToSchedule(TimeSlots),
	length(Variables,TimeSlots),
	mkVariables(Blocks,Variables,ND,NR,NT).
	
mkVariables([],[],_,_,_).
mkVariables([],[_|_],_,_,_) :- 
  message(['mkVariables Fehler Vorberechnete Anzahl der Zeiteinheiten falsch']).
mkVariables([BH|BT],Variables,ND,NR,NT):-
  initVarsForBlock(BH,Variables,VariableRest,ND,NR,NT),
  mkVariables(BT,VariableRest,ND,NR,NT).
  
initVarsForBlock(block(BID,Name,Structure),Variables,VariableRest,ND,NR,NT) :-
     (
      Structure=single -> Sum1 is 1 ;
        Structure=singleNoCluster -> Sum1 is 1 ; 
          Structure=block2 -> Sum1 is 2 ;
            Structure=block2NoCluster -> Sum1 is 2 ;
              Structure=block3 -> Sum1 is 3;
                Structure=block3NoCluster -> Sum1 is 3 ; (=(Sum1,-1),message(['initVarsForBlock/5 Clause 1 -> unbekannte Blockung']))
     ), 
     not(=(Sum1,-1)),
     initVarsForBlock(block(BID,Name,Structure),Variables,VariableRest,ND,NR,NT,Sum1).

initVarsForBlock(_,VT,VT,_,_,_,0).
initVarsForBlock(block(BID,VName,Struct),[VH|VT],VariableRest,ND,NR,NT,Anzahl) :-
    VH=dtr(D,T,R,BID),
    ND0 is ND-1, NT0 is NT-1, NR0 is NR-1,
    D::0..ND0,
    T::0..NT0,
    R::0..NR0,
    N is Anzahl - 1,
    initVarsForBlock(block(BID,VName,Struct),VT,VariableRest,ND,NR,NT,N).


/*************************************************************************/
%
% mkXVariables/2(+VarList,-XVars)
%
% Nimmt die Liste der dtr/4-Constraintvariablenstrukturen entgegen, die den
% Suchraum bildet, und erstellt ber allen dtr(Tag,Zeit,Raum,_) Constraint-
% Variablen ein Constraint 
%
%       NoOfDaysAndTimes * Raum + NoOfTimes * Tag + Zeit #= X
%
% Die jeweils neuen Constraintvariablen X werden in der Ergebnisliste 
% gesammelt. ber dieser Ergebnisliste wendet die Proc. startTimetabling
% das alldifferent/1 - Listen-Constraint aus der fd-Library an. Dies fhrt
% dazu, da alle X paarweise verschieden sein mssen. Dies stellt zusammen 
% mit dem oben angegebenen linearen Constraints sicher das in keinem Raum
% zur selben Zeit mehr als eine Vorlesung stattfindet.
%
% +VarList : Liste der dtr/4-Constraintvariablenstrukturen des Problems
% -XVars   : Liste von Constraintvariablen. Siehe obige Beschreibung.
%
%*************************************************************************/
mkXVariables([],[]).
mkXVariables([dtr(D,T,R,_)|Vars],[X|XVars]) :- 
	noOfDaysAndTimes(NoOfDaysAndTimes),    /* NoOfDays * NoOfTimes */
	noOfTimes(NoOfTimes),
	NoOfDaysAndTimes * R + NoOfTimes * D + T #= X,
	mkXVariables(Vars,XVars).





/*********************************************************************************************/
/* Nummeriert die Namen der Liste "NameList" durch -> NameNumerList ::= nn(Name,LfdNr) *     */
/*********************************************************************************************/
nameNumberList(NameList,NameNumberList) :-
	nameNumberList(NameList,0,NameNumberList).
nameNumberList([],_,[]).
nameNumberList([Name | NameList],I,[nn(Name,I) | NameNumberList]) :-
	I1 is I + 1,
	nameNumberList(NameList,I1,NameNumberList).


name2Number(Name,NameNumberList,Number) :-
	member(nn(Name,Number),NameNumberList).

names2Numbers(Names,NameNumberList,Numbers) :-
	findall(I,(member(Name,Names),member(nn(Name,I),NameNumberList)),Numbers).

numbers2Names(Numbers,NameNumberList,Names) :-
	findall(Name,(member(I,Numbers),member(nn(Name,I),NameNumberList)),Names).


/*courseNumber(CourseName,CourseNumber) :-
	courseNumbers(CourseNumbers),
	name2Number(CourseName,CourseNumbers,CourseNumber),
	!.
courseNumber(CourseName,_) :-
	fail.*/

timeNumber(TimeName,TimeNumber) :-
	timeNumbers(TimeNumbers),
	name2Number(TimeName,TimeNumbers,TimeNumber),
	!.
/*timeNumber(TimeName,ID) :-
	message(['unknown time ID : ',TimeName,' ',ID]),
	fail.*/

dayNumber(DayName,DayNumber) :-
	dayNumbers(DayNumbers),
	name2Number(DayName,DayNumbers,DayNumber),
	!.
/*dayNumber(DayName,ID) :-
	message(['unknown day ID : ',DayName,' ',ID]),
	fail.*/

roomNumber(RoomName,RoomNumber) :-
	roomNumbers(RoomNumbers),
	name2Number(RoomName,RoomNumbers,RoomNumber),
	!.
/*roomNumber(RoomName,ID) :-
	message(['unknown room ID: ',RoomName,' ',ID]),
	fail.*/


/* allGroups(+Group,-AllSubGroups) */
allGroups([],[]).
allGroups([Group | Groups], NewGroups) :-
	allGroups0(Group,Groups1),
	allGroups(Groups,Groups2),
	append(Groups1,Groups2,NewGroups).

allGroups0(Group,Groups) :-
	subgroups(Group,Groups1),
	!,
	allGroups(Groups1,Groups).

allGroups0(Group,[Group]).

groupSize(Group,Size) :-		/* size given directly				*/
	groupsize(Group,Size),
	!.

groupSize(Group,Size) :-		/* size computed by summing up subgroups	*/
	subgroups(Group,Groups),
	sumGroupSize(Groups,Size),
	!.

groupSize(Group,0) :-
	singlePerson(Group),
	!.

groupSize(Group,0) :-
	message(['no size given for group : ',Group]).

sumGroupSize([],0).
sumGroupSize([Group | Groups],Size) :-
	groupSize(Group,Size1),
	sumGroupSize(Groups,Size2),
	Size is Size1 + Size2.

dayNumberList(DayNumbers,DayBlockFactors) :-
	findall(blockFactor(Tag,BlockFactor),tag(Tag,BlockFactor),TagList), DayBlockFactors=TagList,
	findall(Tag,member(blockFactor(Tag,_),TagList),DayList),
	nameNumberList(DayList,DayNumbers).

timeNumberList(TimeNumbers,TimeBlockFactors) :-
	findall(blockFactor(Zeit,BlockFactor),zeit(Zeit,BlockFactor),ZeitList), TimeBlockFactors=ZeitList,
	findall(Zeit,member(blockFactor(Zeit,_),ZeitList),TimeList),
	nameNumberList(TimeList,TimeNumbers).

/* Liefert die Liste "RoomNumbers" aller nn(RaumName,RaumNr)  */
roomNumberList(RoomNumbers,RoomBlockFactors) :-
	findall(blockFactor(Raum,BlockFactor),raum(Raum,_,BlockFactor),RaumList), RoomBlockFactors=RaumList,
	findall(Raum,member(blockFactor(Raum,_),RaumList),RoomList),
	nameNumberList(RoomList,RoomNumbers).

roomHasResource(groesse(Size),Room) :-
	!,
	raum(Room,Resources,_),
	member(groesse(RoomSize),Resources),
	RoomSize >= Size.
roomHasResource(oder(ResListe),Room) :-
        !,roomHasAtLeastOneResource(ResListe,Room).    
roomHasResource(Resource,Room) :-
	raum(Room,Resources,_),
	member(Resource,Resources).
	
roomHasAtLeastOneResource([],_):- fail.
roomHasAtLeastOneResource([H|_],Room) :-
    roomHasResource(H,Room),!.
roomHasAtLeastOneResource([_|T],Room) :-
    roomHasAtLeastOneResource(T,Room).

roomsHaveResource(Resource,Rooms) :-
	findall(Room,roomHasResource(Resource,Room),Rooms).

roomsHaveResources([],Rooms) :-
	findall(Room,raum(Room,_,_),Rooms).
roomsHaveResources([Resouce | Resources],Rooms) :-
	roomsHaveResource(Resouce,Rooms1),
	roomsHaveResources(Resources,Rooms2),
	findall(Room,(member(Room,Rooms1),member(Room,Rooms2)),Rooms).


/* Berechnet die Gesamtzahl der zu terminierenden Doppelstunden */
/* (Benutzt block\3 Liste) */
timesToSchedule(T) :-
    blockNumbers(BlockNumbers),
    timesToSchedule(BlockNumbers,T).

timesToSchedule([],0).
timesToSchedule([block(_,_,H)|T],Sum) :-
    timesToSchedule(T,Sum0),
    blockStructureSize(H,Sum1),
    Sum is Sum0 + Sum1.



blockStructureSize(Structure,Size) :-
    (
      Structure=single -> Size is 1 ;
        Structure=singleNoCluster -> Size is 1 ; 
          Structure=block2 -> Size is 2 ;
            Structure=block2NoCluster -> Size is 2 ;
              Structure=block3 -> Size is 3;
                Structure=block3NoCluster -> Size is 3 ; (Size is 0,message(['illegale Blockstruktur ',Structure]))
    ).
    
removeDoubles([],[]).
removeDoubles([H|T],Res) :-
    member(H,T),!, 
    removeDoubles(T,Res).
removeDoubles([H|T],Res) :-
    removeDoubles(T,Res0),
    append([H],Res0,Res).





